function [x, y_W, y_TV, out] = pdNCG_TV_W(opts)
%PDNCG pdNCG solver for the l1-regularized problem
%
% min tau_TV*||x||_TV + tau_W*||W^Tx||_1 + 0.5||Ax-b||_2^2.

%% Initialize output.
out.itr  = 1; % pdNCG iterations.
out.f = sparse(opts.maxpdNCGitr,1); % Objective values at every iteration.
out.t = sparse(opts.maxpdNCGitr,1); % time passed at every iteration.

%% Initialize data.
stagnation = 0;

% PCG data structure that carries everything the call to the
% preconditioned CG method needs.
PCG = [];

%% Check if A'*A = I, i.e. denoising.
tmp = rand(opts.n1*opts.n2,1);
if norm(opts.A'*(opts.A*tmp)-tmp,1)/norm(tmp,1) < 1e-15;
    opts.reg = 1.0e-16;
    PCG.rho = 1;
else
    PCG.rho = 0.5;
end
clear tmp;

nwtn = zeros(opts.n,1);

x          = opts.x0;
Wt_TV_re_x = opts.Wt_TV_re*x;
Wt_TV_im_x = opts.Wt_TV_im*x;
Wt_TV_x    = Wt_TV_re_x +1i*Wt_TV_im_x;
Wt_W_x     = opts.W'*x;

D_TV        = 1./sqrt(opts.mu^2 + conj(Wt_TV_x).*Wt_TV_x);
DWt_TV_re_x = D_TV.*Wt_TV_re_x;
DWt_TV_im_x = D_TV.*Wt_TV_im_x;

D_W     = 1./sqrt(opts.mu^2 + Wt_W_x.^2);
DWt_W_x = D_W.*Wt_W_x;

y_TV_re    = real(opts.y_TV0);
y_TV_im    = imag(opts.y_TV0);
y_W        = opts.y_W0;

% Calculate the approximate Hessian of pseudo-Huber function for TV.
scale_TV_1  = D_TV.*(1 - DWt_TV_re_x.*y_TV_re);
scale_TV_2  = D_TV.*(DWt_TV_im_x.*y_TV_re);
scale_TV_3  = D_TV.*(DWt_TV_re_x.*y_TV_im);
scale_TV_4  = D_TV.*(1 - DWt_TV_im_x.*y_TV_im);
H_TV        = @(q) hesvec_TV(q,opts.Wt_TV_re,opts.Wt_TV_im,scale_TV_1,...
                             scale_TV_2,scale_TV_3,scale_TV_4);
                         
% Calculate the approximate Hessian of pseudo-Huber function for l1-analysis.                         
scale_W = D_W.*(1 - DWt_W_x.*y_W);
H_W     = @(q) opts.W*(scale_W.*(opts.W'*q));

% Calculate gradient at x0.
g_ls    = opts.A.'*(opts.A*x-opts.b);
g_TV_ph = opts.Wt_TV_re'*DWt_TV_re_x + opts.Wt_TV_im'*DWt_TV_im_x;
g_W_ph  = opts.W*DWt_W_x;
g       = opts.tau_TV.*g_TV_ph + opts.tau_W*g_W_ph + g_ls;

% Fidelity and regularization functions.
f_ls    = @(q) 0.5*norm(opts.A*q-opts.b)^2;
f_huber = @(q) opts.tau_TV*sum(sqrt(opts.mu^2 + ...
               abs(opts.Wt_TV_re*q + 1i*opts.Wt_TV_im*q).^2) - opts.mu) + ...
               opts.tau_W*sum(sqrt(opts.mu^2 + (opts.W'*q).^2) - opts.mu);
               
%% Print
if opts.verbose >= 1
    fprintf('%5s | %5s    | %5s | %5s |  %5s  \n',...
        'pdNCG itr', 'a', '|d|_x/|x|','PCG itr', 'PCG rs');
end

%% Start Newton conjugate gradients.
while 1        
    % Define regularized second-order derivative.
    Hessian = @(q) opts.tau_TV.*H_TV(q) + opts.tau_W.*H_W(q) + ...
                   opts.A'*(opts.A*q) + opts.reg*q;

    % Call user function.
    if isfield(opts, {'user_fcn'}) 
        opts.user_fcn(opts,x,[y_W;y_TV_re + 1i*y_TV_im],...
                          @(q) f_huber(q) + f_ls(q),g,Hessian,nwtn);
    end                 
               
    % Calculate Newton direction.  
    [nwtn, rlrs, pcgi] = PCG_wrapper(PCG, Hessian, g, opts);        
    
    normN_scaled = -(nwtn'*g + opts.reg*norm(nwtn)^2)/norm(x);    
        
    Wt_TV_re_nwtn = opts.Wt_TV_re*nwtn;
    Wt_TV_im_nwtn = opts.Wt_TV_im*nwtn;
    dy_TV_re = scale_TV_1.*Wt_TV_re_nwtn  + scale_TV_2.*Wt_TV_im_nwtn ...
               -(y_TV_re - DWt_TV_re_x);
    dy_TV_im = scale_TV_4.*Wt_TV_im_nwtn  + scale_TV_3.*Wt_TV_re_nwtn ...
               -(y_TV_im - DWt_TV_im_x);
        
    y_TV_tilde_re = y_TV_re + dy_TV_re;
    y_TV_tilde_im = y_TV_im + dy_TV_im;
    y_TV_tilde    = y_TV_tilde_re + 1i*y_TV_tilde_im;
    y_TV_bar      = min(1./abs(y_TV_tilde),1).*y_TV_tilde;    
    y_TV_re       = real(y_TV_bar);
    y_TV_im       = imag(y_TV_bar);
    
    Wt_W_nwtn = opts.W'*nwtn;
    dy_W      = scale_W.*Wt_W_nwtn - (y_W - DWt_W_x);
        
    y_tilde_W = y_W + dy_W;
    y_W       = sign(y_tilde_W).*min(abs(y_tilde_W),1);
    
    % Calculate stepsize and update variables.
    [x, a, f_new] = simple_bt(x, nwtn, f_ls, f_huber, normN_scaled, opts);
    
    % Store objective function and corresponding time.
    out.f(out.itr) = f_new;
    out.t(out.itr) = toc; 
    
    % The PCG methods needs to know the stepsize.
    PCG.a = a;
        
    % Printing info.
    if opts.verbose >= 1
        fprintf('  %4d    | %1.2e |  %8s | %4d    | %1.2e \n',...
            out.itr, a, num2str(normN_scaled,'%1.2e\n'), pcgi, rlrs);
    end
        
    % Calculate the approximate Hessian of pseudo-Huber function for TV.
    Wt_TV_re_x = opts.Wt_TV_re*x;
    Wt_TV_im_x = opts.Wt_TV_im*x;
    Wt_TV_x    = Wt_TV_re_x +1i*Wt_TV_im_x;
    
    D_TV        = 1./sqrt(opts.mu^2 + conj(Wt_TV_x).*Wt_TV_x);
    DWt_TV_re_x = D_TV.*Wt_TV_re_x;
    DWt_TV_im_x = D_TV.*Wt_TV_im_x;
    scale_TV_1  = D_TV.*(1 - DWt_TV_re_x.*y_TV_re);
    scale_TV_2  = D_TV.*(DWt_TV_im_x.*y_TV_re);
    scale_TV_3  = D_TV.*(DWt_TV_re_x.*y_TV_im);
    scale_TV_4  = D_TV.*(1 - DWt_TV_im_x.*y_TV_im);
    H_TV        = @(q) hesvec_TV(q,opts.Wt_TV_re,opts.Wt_TV_im,...
                                 scale_TV_1,scale_TV_2,scale_TV_3,scale_TV_4);
                             
    % Calculate the approximate Hessian of pseudo-Huber function for l1-analysis.  
    Wt_W_x = opts.W'*x;
    
    D_W     = 1./sqrt(opts.mu^2 + Wt_W_x.^2);
    DWt_W_x = D_W.*Wt_W_x;
    scale_W = D_W.*(1 - DWt_W_x.*y_W);
    H_W     = @(q) opts.W*(scale_W.*(opts.W'*q));

    % Calculate gradient at x.
    g_ls    = opts.A.'*(opts.A*x-opts.b);
    g_TV_ph = opts.Wt_TV_re'*DWt_TV_re_x + opts.Wt_TV_im'*DWt_TV_im_x;
    g_W_ph  = opts.W*DWt_W_x;
    g       = opts.tau_TV.*g_TV_ph + opts.tau_W*g_W_ph + g_ls;
    
    if isfield(opts, {'user_stop'}) 
        if opts.user_stop(opts,x,[y_W;y_TV_re +1i*y_TV_im],...
                                  @(q) f_huber(q) + f_ls(q),g,Hessian,nwtn)
            out.message = 'Converged';
            break;
        end
    else
        if normN_scaled <= opts.tol
            out.message = 'Converged';
            break; 
        end         
    end
    
    % Maximum iterations reached.
    if out.itr > opts.maxpdNCGitr
        stagnation = 1;
    end
    
    if stagnation == 1
       out.message = 'Maximum iterations of pdNCG reached';
       fprintf('Maximum iterations of pdNCG reached, termination.\n'); 
       fprintf('Possible fixes: \n');
       fprintf('a) increase opts.maxpdNCGitr. Performs more pdNCG iterations.\n');
       fprintf('b) decrease opts.tolpcg. Calculates better directions. \n');
       fprintf('c) increase opts.mxitrpcg. \n');
       break;
    end 
    
    % Update iteration counter.
    out.itr = out.itr + 1;
end
%% End of primal-dual Newton conjugate gradients.

out.f = out.f(1:out.itr);
out.t = out.t(1:out.itr);

% Return final complex dual variable.
y_TV = y_TV_re + 1i*y_TV_im;

% Call user function.
if isfield(opts, {'user_fcn'}) 
        opts.user_fcn(opts,x,[y_W;y_TV_re + 1i*y_TV_im],...
                          @(q) f_huber(q) + f_ls(q),g,Hessian,nwtn);
end  